﻿namespace HDLNoCGen
{        
    /// <summary>
    /// Перечисление типов графов.
    /// </summary>
    public enum GraphType
    {
        /// <summary>
        /// Циркулянт.
        /// </summary>
        Circulant,
        /// <summary>
        /// Сетевой граф.
        /// </summary>
        Mesh,
        /// <summary>
        /// Тор.
        /// </summary>
        Torus
    }

    /// <summary>
    /// Набор опций для роутера.
    /// </summary>
    public struct RouterOptions
    {
        /// <summary>
        /// Перечисление типов арбитров.
        /// </summary>
        public enum Arbiter_types
        {
            Round_Robin
        };

        /// <summary>
        /// Перечисление типов очередей.
        /// </summary>
        public enum Queue_types
        {
            Index,
            Pointer,
            Line
        };

        /// <summary>
        /// Выбор позиции очереди.
        /// </summary>
        public enum Queue_position
        {
            Front,
            Rear
        };

        /// <summary>
        /// Выбор позиции очереди.
        /// </summary>
        public enum Algorithm
        {
            XY,
            GA
        };
    }

    /// <summary>
    /// Структура типа текущего роутера.
    /// </summary>
    public struct router_options
    {
        public RouterOptions.Arbiter_types arbiter_type { get; set; }
        public RouterOptions.Queue_types queue_type { get; set; }
        public RouterOptions.Queue_position queue_position { get; set; }
        public RouterOptions.Algorithm algorithm { get; set; }

        public int packet_width { get; set; }
        public int info_width { get; set; }
        public int queue_length { get; set; }
    }

    public class GraphDeserializer
    {
        public int node_count { get; set; }

        public List<int> generators { get; set; }

        public router_options parameters { get; set; }

        public GraphType graphId { get; set; }
    }


    /// <summary>
    /// Класс, представляющий граф.
    /// </summary>
    public abstract class Graph
    {

        /// <summary>
        /// Сообщение об ошибке.
        /// </summary>
        protected string error_message { get; set; }

        /// <summary>
        /// Даиметр графа.
        /// </summary>
        protected int diameter { get; set; }

        /// <summary>
        /// Матрица смежности графа.
        /// </summary>
        protected int[,] matr_smej { get; set; }

        /// <summary>
        /// Среднее расстояние между узлами графа.
        /// </summary>
        protected double average_distance { get; set; }

        /// <summary>
        /// Минимальные пути по алгоритму Дейкстры.
        /// </summary>
        protected List<List<int>> min_ways_routing_deikstra { get; set; }

        /// <summary>
        /// Эффективность алгоритма Дейкстры.
        /// </summary>
        protected double efficiency_deikstra { get; set; }

        /// <summary>
        /// Указывает, был ли создан граф.
        /// </summary>
        protected bool is_created { get; set; }

        /// <summary>
        /// Количество узлов в графе.
        /// </summary>
        public int node_count { get; set; }

        /// <summary>
        /// Список генераторов графа.
        /// </summary>
        public List<int> generators { get; set; }

        public router_options parameters { get; set; }

        public GraphType graphId { get; set; }

        /// <summary>
        /// Конструктор графа.
        /// </summary>
        public Graph()
        {
            this.error_message = "";

            this.is_created = false;
            this.generators = null;
            this.node_count = 0;
            this.efficiency_deikstra = 1;
        }

        public Graph(GraphDeserializer graph_des)
        {
            this.parameters = graph_des.parameters;
            this.generators = graph_des.generators;
            this.node_count = graph_des.node_count;
            this.graphId = graph_des.graphId;
        }

        public abstract string getSignature();

        protected abstract void SetNodeCountFromGenerators(List<int> generators);

        public void InstallGenerators(List<int> generators, CancellationToken token)
        {
            SetNodeCountFromGenerators(generators);
            generators.Sort();
            for (int i = 0; i < generators.Count - Convert.ToInt32(this.graphId == GraphType.Circulant); i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                generators[i] = generators[i] % this.node_count;
            }

            generators.Remove(0);
            this.generators = generators;

            this.is_created = true;

            this.Calculate_diameter(token);
            this.Generate_deikstra_routing(token);

            GC.Collect();
        }

        /// <summary>
        /// Создает матрицу смежности (абстрактный метод).
        /// </summary>
        protected abstract void Create_matr_smej();

        /// <summary>
        /// Вычисляет эффективность алгоритма Дейкстры.
        /// </summary>
        /// <returns>Эффективность алгоритма Дейкстры.</returns>
        public double Calculate_efficiency_deikstra()
        {
            this.efficiency_deikstra = 1.0;

            return this.efficiency_deikstra;
        }

        /// <summary>
        /// Вычисляет диаметр графа на основе маршрутов, найденных алгоритмом Дейкстры.
        /// </summary>
        /// <param name="token">Токен для отмены операции.</param>
        /// <returns>Диаметр графа или -1, если маршруты не найдены.</returns>
        public int Calculate_diameter(CancellationToken token)
        {
            if (this.min_ways_routing_deikstra == null)
                return -1;
            this.diameter = 0;
            for (int i = 0; i < this.min_ways_routing_deikstra.Count; i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                if (this.min_ways_routing_deikstra[i].Count > this.diameter)
                {
                    this.diameter = this.min_ways_routing_deikstra[i].Count;
                }
            }
            --this.diameter;  // Уменьшаем диаметр на 1, так как считаем количество рёбер

            return this.diameter;
        }

        /// <summary>
        /// Генерирует маршруты с использованием алгоритма Дейкстры.
        /// </summary>
        /// <param name="token">Токен отмены для управления отменой операции.</param>
        /// <returns>Список кратчайших маршрутов между узлами графа.</returns>
        /// <exception cref="OperationCanceledException">Выход, если операция была отменена.</exception>
        public List<List<int>> Generate_deikstra_routing(CancellationToken token)
        {
            // Создаем матрицу смежности графа
            Create_matr_smej();

            // Инициализируем список для хранения минимальных маршрутов
            min_ways_routing_deikstra = new List<List<int>>();

            // Проходим по каждому узлу графа
            for (int m = 0; m < node_count; m++)
            {
                // Массивы для хранения расстояний и посещенных узлов
                int[] d = new int[node_count]; // Расстояния от начального узла
                int[] v = new int[node_count]; // Статус посещения узлов
                int minindex;  // Индекс узла с минимальным расстоянием
                int min;  // Минимальное расстояние
                int temp;  // Временная переменная для вычислений
                List<int[]> ways;  // Список для хранения путей

                // Инициализация массивов
                for (int i = 0; i < node_count; i++)
                {
                    if (token.IsCancellationRequested)
                    {
                        token.ThrowIfCancellationRequested();
                    }
                    d[i] = 10000;  // Устанавливаем максимальное значение расстояния
                    v[i] = 1;  // Устанавливаем, что узел еще не посещен
                }
                d[m] = 0;  // Расстояние до начального узла равно 0

                // Шаг алгоритма
                do
                {
                    minindex = 10000;
                    min = 10000;
                    for (int i = 0; i < node_count; i++)
                    { // Если вершину ещё не обошли и вес меньше min
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        if ((v[i] == 1) && (d[i] < min))
                        { // Переприсваиваем значения
                            min = d[i];
                            minindex = i;
                        }
                    }

                    // Добавляем найденный минимальный вес к текущему весу вершины
                    // и сравниваем с текущим минимальным весом вершины
                    if (minindex != 10000)
                    {
                        for (int i = 0; i < node_count; i++)
                        {
                            if (matr_smej[minindex, i] > 0)
                            {
                                temp = min + matr_smej[minindex, i];
                                if (temp < d[i])
                                {
                                    d[i] = temp;
                                }
                            }
                        }
                        v[minindex] = 0;
                    }
                } while (minindex < 10000);
                //конец алгоритма Дейкстра


                //восстанавливаем пути
                ways = new List<int[]>();
                for (int i = 0; i < node_count; i++)
                {
                    int[] ver;
                    ver = new int[node_count];
                    for (int j = 0; j < node_count; j++)
                    {
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        ver[j] = -1;
                    }
                    int end = i;
                    ver[0] = end;
                    int k = 1;
                    int weight = d[end];

                    while (end != m)//Пока не окажимся в начальной вершине
                    {
                        for (int j = 0; j < node_count; j++)
                        {
                            if (token.IsCancellationRequested)
                            {
                                token.ThrowIfCancellationRequested();
                            }
                            if (matr_smej[end, j] > 0)//Если вершины смежны
                            {
                                temp = weight - matr_smej[end, j];//Предсказываем вес вершины
                                if (temp == d[j])//Если вес вершины совпал с предсказанным
                                {
                                    weight = temp;
                                    end = j;//переходим в вершину
                                    ver[k] = j;//записваем переход
                                    k++;
                                    break;
                                }
                            }
                        }
                    }
                    ways.Add(ver);//Записываем путь
                }

                for (int i = 0; i < ways.Count; i++)//Расшифровываем пути
                {
                    List<int> buff = new List<int>();  // Временный список для хранения маршрута
                    for (int j = node_count - 1; j > -1; j--)
                    {
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        if (ways[i][j] != -1)//Если переход не пустой
                        {
                            buff.Add(ways[i][j]);  // Добавляем вершину в маршрут
                        }
                    }
                    min_ways_routing_deikstra.Add(buff);  // Добавляем маршрут в общий список маршрутов
                }
            }

            return this.min_ways_routing_deikstra;  // Возвращаем список минимальных маршрутов
        }

        /// <summary>
        /// Создает экземпляр графа в зависимости от указанного типа графа.
        /// </summary>
        /// <param name="graphType">Тип графа.</param>
        /// <param name="buffer">Строка с параметрами для создания графа.</param>
        /// <param name="token">Токен отмены.</param>
        /// <returns>Созданный экземпляр графа.</returns>
        /// <exception cref="Exception">Выход, если тип графа не поддерживается.</exception>
        public Graph CreateGraphInstance(GraphType graphType, string buffer, CancellationToken token)
        {
            switch (graphType)
            {
                case GraphType.Circulant:
                    return new GraphCirculant(buffer, token);
                case GraphType.Mesh:
                    return new GraphMesh(buffer, token);
                case GraphType.Torus:
                    return new GraphTorus(buffer, token);
                default:
                    throw new Exception("Unsupported graph type");
            }
        }

        /// <summary>
        /// Создает граф на основе входного буфера и возвращает его.
        /// </summary>
        /// <param name="buffer">Строка, содержащая информацию о графе.</param>
        /// <param name="token">Токен для отмены операции.</param>
        /// <returns>Созданный экземпляр графа.</returns>
        /// <exception cref="Exception">Выход, если тип топологии графа некорректен.</exception>
        public Graph CreateGraphFromBuffer(string buffer, CancellationToken token)
        {
            GraphType graphType;

            // Определение типа графа на основе ввода
            switch (buffer.Split(' ')[0])
            {
                case "c":
                case "C":
                case "с":
                case "С":
                    graphType = GraphType.Circulant;
                    break;
                case "m":
                case "M":
                    graphType = GraphType.Mesh;
                    break;
                case "t":
                case "T":
                    graphType = GraphType.Torus;
                    break;
                default:
                    throw new Exception("Incorrect topology");
            }

            return CreateGraphInstance(graphType, buffer, token);
        }

        /// <summary>
        /// Виртуальная функция для генерации части проекта, отвечающей за функционирование роутера.
        /// </summary>
        /// <param name="project_path">Путь к корневой папке проекта.</param>
        /// <param name="parametrs">Параметры роутера.</param>
        public virtual void createRouter(string project_path, router_options parametrs)
        {
            if (Directory.Exists(project_path))
            {
                FileSystemWatcher watcher = new FileSystemWatcher();
                AutoResetEvent waitDeleteHandle = new AutoResetEvent(false);

                watcher.Path = project_path;
                watcher.EnableRaisingEvents = true;
                watcher.Deleted += new FileSystemEventHandler((sender, e) =>
                {
                    Directory.CreateDirectory(project_path);
                    Directory.CreateDirectory(project_path + "\\src");
                    Directory.CreateDirectory(project_path + "\\inc");
                    waitDeleteHandle.Set();
                });

                Directory.Delete(project_path, true);
                waitDeleteHandle.WaitOne();
            }
            else
            {
                Directory.CreateDirectory(project_path);
                Directory.CreateDirectory(project_path + "\\src");
                Directory.CreateDirectory(project_path + "\\inc");
            }

            File.Copy(descipher_router(parametrs.arbiter_type), project_path + "\\src\\arbiter.sv");
            File.Copy(descipher_router(parametrs.algorithm), project_path + "\\src\\algorithm.sv");


            string queue_path = descipher_router(parametrs.queue_type);
            int queue_length = (int)parametrs.queue_length;

            switch (parametrs.queue_type)
            {
                case RouterOptions.Queue_types.Line:
                    string queue_text = File.ReadAllText(queue_path);
                    string casez_text = "";

                    #region casegenerator

                    casez_text += "            " + "'b0" + new string('?', queue_length) + "1:\n";
                    casez_text += "            " +  "begin\n";
                    casez_text += "            " +  "    queue_buffers[0] <= data_in;\n";
                    casez_text += "            " +  "end\n";

                    for (int i = 2; i < queue_length * 2; i++)
                    {
                        casez_text += "            " +  "'b" + new string('1', i / 2) + "0" + new string('?', queue_length - (i / 2) - 1) + $"{Convert.ToInt16(i % 2 == 0)}" + "1:\n";
                        casez_text += "            " +  "begin\n";

                        for (int j = 1; (i % 2 == 0) && (j < i / 2); j++)
                        {
                            casez_text += "            " +  $"    queue_buffers[{j - 1}] <= queue_buffers[{j}];\n";
                        }
                        casez_text += "            " +  $"    queue_buffers[{i / 2 - Convert.ToInt16(i % 2 == 0)}] <= data_in;\n";

                        casez_text += "            " +  "end\n";
                    }

                    casez_text += "            " +  "'b" + new string('1', queue_length + 2) + ":\n";
                    casez_text += "            " +  "begin\n";
                    for (int i = 1; i < queue_length; i++)
                    {
                        casez_text += "            " +  $"    queue_buffers[{i - 1}] <= queue_buffers[{i}];\n";
                    }
                    casez_text += "            " +  $"    queue_buffers[{queue_length - 1}] <= 0;\n";
                    casez_text += "            " +  "end\n";

                    casez_text += "            " +  "\n";


                    for (int i = 1; i <= queue_length; i++)
                    {
                        casez_text += "            " +  "'b" + new string('1', i) + new string('0', Convert.ToInt32(i != queue_length)) + new string('?', Int32.Max(queue_length - i - 1, 0)) + "10:\n";
                        casez_text += "            " +  "begin\n";

                        for (int j = 1; j < i; j++)
                        {
                            casez_text += "            " +  $"    queue_buffers[{j - 1}] <= queue_buffers[{j}];\n";
                        }

                        casez_text += "            " +  $"    queue_buffers[{i - 1}] <= 0;\n";
                        casez_text += "            " +  "end\n\n";
                    }

                    casez_text += "            " +  "default:\n";
                    casez_text += "            " +  "begin\n";
                    casez_text += "            " +  "end";

                    #endregion

                    queue_text = String.Format(queue_text, casez_text);
                    File.WriteAllText(project_path + "\\src\\queue.sv", queue_text);

                    break;
                case RouterOptions.Queue_types.Index:
                case RouterOptions.Queue_types.Pointer:
                    File.Copy(queue_path, project_path + "\\src\\queue.sv");
                    break;
                default:
                    throw new NotImplementedException($"Queue type {parametrs.queue_type} was not implemented into this switch statement.");
            }

        }

        /// <summary>
        /// Виртуальная функция для генерации части проекта, отвечающей за файл топологии.
        /// </summary>
        /// <param name="project_path">Путь к корневой папке проекта.</param>
        public virtual void createNoC(string project_path, router_options parameters) 
        {
            Directory.CreateDirectory(project_path + "\\noc");
            Directory.CreateDirectory(project_path + "\\tb");
        }

        /// <summary>
        /// Получает минимальную длину числа в битах.
        /// </summary>
        /// <returns>Размер координат.</returns>
        protected abstract int GetCordinatesSize(); //Returns how many bits are required to store coordinates

        public int get_bits(int num) 
        { 
            int bits = 1;
            int counter = 2;
            while (counter < num) 
            {
                ++bits;
                counter *= 2;
            }

            return bits;
        }

        /// <summary>
        /// Возвращает относительный путь до запрошенного файла.
        /// </summary>
        /// <param name="obj">Enum запрашиваемого объекта.</param>
        protected string descipher_router(RouterOptions.Arbiter_types obj)
        {
            string response = "..\\..\\..\\verilog-primitives\\src\\arb\\";
            switch (obj)
            {
                case RouterOptions.Arbiter_types.Round_Robin:
                    response += "RR_";
                    break;
                default:
                    throw new NotImplementedException($"Arbiter {obj} was not implemented into this switch statement.");
            }
            return response + "arbiter.sv";
        }

        /// <summary>
        /// Возвращает относительный путь до запрошенного файла.
        /// </summary>
        /// <param name="obj">Enum запрашиваемого объекта.</param>
        protected string descipher_router(RouterOptions.Queue_types obj)
        {
            string response = "..\\..\\..\\verilog-primitives\\src\\bfr\\";
            switch (obj)
            {
                case RouterOptions.Queue_types.Index:
                    response += "index_";
                    break;
                case RouterOptions.Queue_types.Pointer:
                    response += "pointer_";
                    break;
                case RouterOptions.Queue_types.Line:
                    response += "line_";
                    break;
                default:
                    throw new NotImplementedException($"Queue type {obj} was not implemented into this switch statement.");
            }
            return response + "queue.sv";
        }
        /// <summary>
        /// Возвращает относительный путь до запрошенного файла.
        /// </summary>
        /// <param name="obj">Enum запрашиваемого объекта.</param>
        protected abstract string descipher_router(RouterOptions.Algorithm obj);

        /// <summary>
        /// Возвращает относительный путь до запрошенного файла.
        /// </summary>
        /// <param name="obj">Enum запрашиваемого объекта.</param>
        protected string descipher_router(RouterOptions.Queue_position obj)
        {
            string response = "..\\..\\..\\verilog-primitives\\src\\rtr\\";
            switch (obj)
            {
                case RouterOptions.Queue_position.Front:
                    response += "f_";
                    break;
                case RouterOptions.Queue_position.Rear:
                    response += "r_";
                    break;
                default:
                    throw new NotImplementedException($"Queue position {obj} was not implemented into this switch statement.");
            }
            string path = response + "router.sv";
            return File.ReadAllText(path);
        }


    }



}
