﻿using System.Text;
using System.Text.Json;

namespace HDLNoCGen
{
    /// <summary>
    /// Представляет топологию графа "Тор".
    /// Наследуется от базового класса Graph и реализует интерфейс IRectangleLike.
    /// </summary>
    class GraphTorus : Graph, IRectangleLike
    {
        /// <summary>
        /// Возвращает сигнатуру графа (тип топологии).
        /// </summary>
        /// <returns>Строка, представляющая сигнатуру графа.</returns>
        public override string getSignature()
        {
            return "Torus";
        }

        /// <summary>
        /// Список маршрутов XY для каждой пары узлов.
        /// </summary>
        protected List<List<int>> ways_routing_XY { get; set; }

        /// <summary>
        /// Эффективность маршрутизации XY по сравнению с маршрутизацией Дейкстры.
        /// </summary>
        protected double XY_Efficiency { get; set; }

        /// <summary>
        /// Среднее расстояние при маршрутизации XY.
        /// </summary>
        protected double average_distance_XY { get; set; }

        /// <summary>
        /// Вычисляет эффективность маршрутизации XY на основе сравнения с кратчайшими путями Дейкстры.
        /// </summary>
        /// <param name="token">Токен отмены для прерывания вычислений.</param>
        private void calculateEfficiencyXY(CancellationToken token)
        {
            double deikstra_count = 0;
            double XY_count = 0;

            for (int i = 0; i < this.min_ways_routing_deikstra.Count; i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                deikstra_count += this.min_ways_routing_deikstra[i].Count - 1;
                XY_count += this.ways_routing_XY[i].Count - 1;
            }

            this.XY_Efficiency = Math.Round(deikstra_count / XY_count, 3);
        }

        /// <summary>
        /// Устанавливает количество узлов графа на основе списка генераторов.
        /// </summary>
        /// <param name="generators">Список генераторов (размеры тора).</param>
        protected override void SetNodeCountFromGenerators(List<int> generators)
        {
            this.node_count = generators[0] * generators[1];
        }

        /// <summary>
        /// Возвращает список маршрутов XY.
        /// </summary>
        /// <returns>Список маршрутов XY.</returns>
        public List<List<int>> Get_ways_routing_XY()
        {
            return this.ways_routing_XY;
        }

        /// <summary>
        /// Вычисляет и возвращает эффективность маршрутизации XY.
        /// </summary>
        /// <param name="token">Токен отмены для прерывания вычислений.</param>
        /// <returns>Эффективность маршрутизации XY.</returns>
        public double getXY_Efficiency(CancellationToken token)
        {
            calculateEfficiencyXY(token);
            return this.XY_Efficiency;
        }

        /// <summary>
        /// Конструктор по умолчанию для класса GraphTorus.
        /// Инициализирует граф с типом "Тор".
        /// </summary>
        public GraphTorus() : base()
        {
            this.graphId = GraphType.Torus;
            this.diameter = 0;
            this.matr_smej = null;
        }

        /// <summary>
        /// Конструктор для класса GraphTorus, принимающий объект GraphDeserializer.
        /// </summary>
        /// <param name="graph_des">Объект GraphDeserializer для инициализации графа.</param>
        public GraphTorus(GraphDeserializer graph_des) : base(graph_des)
        {

        }

        /// <summary>
        /// Конструктор для класса GraphTorus, принимающий параметры в виде строки.
        /// </summary>
        /// <param name="parameters">Строка параметров, описывающая размеры тора.</param>
        /// <param name="token">Токен отмены для прерывания создания графа.</param>
        public GraphTorus(string parameters, CancellationToken token)
        {
            this.graphId = GraphType.Torus;

            string[] buffer = parameters.Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
            string[] buffer_topology = buffer[0].Trim().Split(new string[] { "T", "t", ";", ":", ", ", ")", " " }, StringSplitOptions.RemoveEmptyEntries);
            generators = new List<int>();
            for (int i = 0; i < buffer_topology.Length && i < 2; i++)
            {
                generators.Add(Convert.ToInt32(buffer_topology[i]));
            }
            generators.Sort();
            
            if (generators.Count < 2 || generators.Min() < 2)
            {
                throw(new Exception("Ошибка создания топологии"));
            }

            this.node_count = generators[0] * generators[1];

            //string[] buffer = parameters.Split(new string[] { "С(", "C(", ";", ":",  ", ", ")", " " }, StringSplitOptions.RemoveEmptyEntries);
            //this.node_count = Convert.ToInt32(buffer[0]);
            //this.p = Convert.ToInt32(param_2);

            this.is_created = true;
        }

        /// <summary>
        /// Создает матрицу смежности для графа "Тор".
        /// </summary>
        protected override void Create_matr_smej()
        {
            if (generators.Count < 2)
                return;
            this.matr_smej = new int[this.node_count, this.node_count];
            int n = generators[0];
            int m = generators[1];

            for (int i = 0; i < this.node_count; i++)
            {
                if (i % m != m - 1)
                {
                    this.matr_smej[i, i + 1] = 1;
                    this.matr_smej[i + 1, i] = 1;
                }
                if (i % m == 0)
                {
                    this.matr_smej[i, i + m - 1] = 1;
                    this.matr_smej[i + m - 1, i] = 1;
                }
                if (i < (this.node_count - m))
                {
                    this.matr_smej[i, i + m] = 1;
                    this.matr_smej[i + m, i] = 1;
                }
                if (i < m)
                {
                    this.matr_smej[i, i + m * (n - 1)] = 1;
                    this.matr_smej[i + m * (n - 1), i] = 1;
                }
            }
        }

        /// <summary>
        /// Генерирует маршруты XY для всех пар узлов в графе.
        /// </summary>
        /// <param name="token">Токен отмены для прерывания генерации маршрутов.</param>
        /// <returns>Список маршрутов XY для каждой пары узлов. Возвращает null в случае ошибки.</returns>
        public List<List<int>> Generate_XY_routing(CancellationToken token)
        {
            int iterations_count;

            int start_node;
            int end_node;
            List<int> route;

            this.ways_routing_XY = new List<List<int>>();

            try
            {
                if (!this.is_created)
                {
                    throw new Exception("Описание топологии не было задано или задано не полностью");
                }

                for (int i = 0; i < this.node_count; i++)
                {
                    for (int j = 0; j < this.node_count; j++)
                    {
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        iterations_count = 0;
                        start_node = i;
                        end_node = j;
                        route = new List<int>();
                        route.Add(start_node + Settings.Get_node_naming_start_index());
                        while (start_node != end_node)
                        {
                            iterations_count++;

                            start_node = XY_routing(start_node, end_node, generators, node_count);
                            route.Add(start_node + Settings.Get_node_naming_start_index());
                            if (iterations_count > Settings.Get_error_iterations_count())
                            {
                                throw new Exception("Превышено количество итераций моделирования маршрута для алгоритма Simple");
                            }
                        }
                        this.ways_routing_XY.Add(route);
                    }
                }
            }
            catch (Exception ex)
            {
                this.error_message = ex.Message;
                return null;
            }

            return this.ways_routing_XY;
        }

        /// <summary>
        /// Выполняет один шаг маршрутизации XY для перемещения от текущего узла к целевому.
        /// </summary>
        /// <param name="start_node">Индекс текущего узла.</param>
        /// <param name="end_node">Индекс целевого узла.</param>
        /// <param name="generators">Список генераторов (размеры тора).</param>
        /// <param name="node_count">Общее количество узлов в графе.</param>
        /// <returns>Индекс следующего узла в маршруте.</returns>
        private int XY_routing(int start_node, int end_node, List<int> generators, int node_count)
        {
            int start_x = start_node % generators[1];
            int end_x = end_node % generators[1];
            int start_y = start_node / generators[1];
            int end_y = end_node / generators[1];
            int step = 0;

            if (start_x < end_x)
            {
                if (end_x - start_x < generators[1] - end_x + start_x)
                {
                    step = 1;
                }
                else
                {
                    if (start_x == 0)
                    {
                        step = generators[1] - 1;
                    }
                    else
                    {
                        step = -1;
                    }
                }
            }
            else if (start_x > end_x)
            {
                if (start_x - end_x < generators[1] - start_x + end_x)
                {
                    step = -1;
                }
                else
                {
                    if (start_x == generators[1] - 1)
                    {
                        step = -generators[1] + 1;
                    }
                    else
                    {
                        step = 1;
                    }
                }
            }
            else
            {
                if (start_y < end_y)
                {
                    if (end_y - start_y < generators[0] - end_y + start_y)
                    {
                        step = generators[1];
                    }
                    else
                    {
                        if (start_y == 0)
                        {
                            step = generators[1] * (generators[0] - 1);
                        }
                        else
                        {
                            step = -generators[1];
                        }
                    }
                }
                else if (start_y > end_y)
                {
                    if (start_y - end_y < generators[0] - start_y + end_y)
                    {
                        step = -generators[1];
                    }
                    else
                    {
                        if (start_y == generators[0] - 1)
                        {
                            step = -generators[1] * (generators[0] - 1);
                        }
                        else
                        {
                            step = generators[1];
                        }
                    }
                }
            }

            return start_node + step;
        }

        /// <summary>
        /// Вычисляет среднее расстояние между узлами при маршрутизации XY.
        /// </summary>
        /// <param name="token">Токен отмены для прерывания вычислений.</param>
        /// <returns>Среднее расстояние при маршрутизации XY.</returns>
        public double Calculate_average_distance_XY(CancellationToken token)
        {
            average_distance_XY = 0;

            for (int i = 0; i < this.ways_routing_XY.Count; i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                average_distance_XY += ways_routing_XY[i].Count;
            }
            average_distance_XY /= ways_routing_XY.Count;

            return average_distance_XY;
        }

        /// <summary>
        /// Возвращает размер координат, необходимый для представления узлов графа.
        /// </summary>
        /// <returns>Размер координат в битах.</returns>
        protected override int GetCordinatesSize()
        {
            int coordinates_size = 0;
            int counter = 1;
            while (counter < generators.Max())
            {
                coordinates_size++;
                counter *= 2;
            }
            return coordinates_size;
        }

        /// <summary>
        /// Преобразует алгоритм маршрутизации в путь к соответствующему файлу Verilog.
        /// </summary>
        /// <param name="obj">Алгоритм маршрутизации.</param>
        /// <returns>Путь к файлу Verilog, реализующему алгоритм маршрутизации.</returns>
        /// <exception cref="NotImplementedException">Выбрасывается, если алгоритм не реализован.</exception>
        protected override string descipher_router(RouterOptions.Algorithm obj)
        {
            string response = "..\\..\\..\\verilog-primitives\\src\\alg\\torus\\";

            switch (obj)
            {
                case RouterOptions.Algorithm.XY:
                    response += "XY.sv";
                    break;
                default:
                    throw new NotImplementedException($"Algorithm {obj} was not implemented into this switch statement.");
            }

            return response;
        }

        /// <summary>
        /// Создает файлы, описывающие маршрутизатор для данной топологии.
        /// </summary>
        /// <param name="project_path">Путь к папке проекта.</param>
        /// <param name="parametrs">Параметры маршрутизатора.</param>
        /// <exception cref="NotImplementedException">Выбрасывается, если комбинация алгоритма и позиции очереди не реализована.</exception>
        public override void createRouter(string project_path, router_options parametrs)
        {
            base.createRouter(project_path, parametrs);

            string router_format = descipher_router(parametrs.queue_position);
            string router_text = "";
            string router_json = File.ReadAllText("..\\..\\..\\router_variants.json");
            var router_variants = JsonSerializer.Deserialize<Dictionary<string, Dictionary<string, Dictionary<string, List<string>>>>>(router_json);

            switch (parametrs.algorithm, parametrs.queue_position)
            {
                case (RouterOptions.Algorithm.XY, RouterOptions.Queue_position.Front):
                    router_text = String.Format(router_format, String.Join("\n    ", router_variants["torus"]["xy"]["front"].Select(s => $"{s}")).Split("c#_splitter"));
                    break;
                case (RouterOptions.Algorithm.XY, RouterOptions.Queue_position.Rear):
                    router_text = String.Format(router_format, String.Join("\n    ", router_variants["torus"]["xy"]["rear"].Select(s => $"{s}")).Split("c#_splitter"));
                    break;
                default:
                    throw new NotImplementedException($"Router format {parametrs.algorithm}, {parametrs.queue_position} was not implemented into this switch statement.");
            }

            File.WriteAllText(project_path + "\\src\\router.sv", router_text);


            string file_to_format = "noc";
            string file_text = "";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                parametrs.packet_width,
                GetCordinatesSize(),
                node_count
                ));

            file_to_format = "noc_XY";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                generators[0], generators[1]
                ));

            file_to_format = "queue";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                parametrs.queue_length, get_bits(parametrs.queue_length)
                ));

            file_to_format = "router";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                5,
                this.get_bits(5)
                ));

        }

        /// <summary>
        /// Создает файлы, описывающие NoC (Network-on-Chip) для данной топологии.
        /// </summary>
        /// <param name="project_path">Путь к папке проекта.</param>
        /// <param name="parameters">Параметры NoC.</param>
        public override void createNoC(string project_path, router_options parameters)
        {
            base.createNoC(project_path, parameters);

            File.Copy("..\\..\\..\\verilog-primitives\\noc\\torus.sv", project_path + "\\noc\\noc.sv");
            File.Copy("..\\..\\..\\verilog-primitives\\sel\\rectangle.sv", project_path + "\\noc\\toplevel.sv");
            File.Copy("..\\..\\..\\verilog-primitives\\tb\\routing\\rectangle.sv", project_path + "\\tb\\tb_routing.sv");

            switch (parameters.queue_position)
            {
                case RouterOptions.Queue_position.Front:
                    File.Copy("..\\..\\..\\verilog-primitives\\tb\\queueing\\front\\rectangle.sv", project_path + "\\tb\\tb_queueing.sv");
                    break;
                case RouterOptions.Queue_position.Rear:
                    File.Copy("..\\..\\..\\verilog-primitives\\tb\\queueing\\rear\\rectangle.sv", project_path + "\\tb\\tb_queueing.sv");
                    break;
            }
        }
    }
}