﻿namespace HDLNoCGen
{
    /// <summary>
    /// Определяет интерфейс для классов, реализующих функциональность генерации маршрутизации в NoC (Network-on-Chip) и расчета эффективности.
    /// </summary>
    /// <remarks>
    /// Этот интерфейс содержит методы для генерации различных типов таблиц маршрутизации,
    /// а также для вычисления эффективности различных алгоритмов маршрутизации.
    /// </remarks>
    internal interface IRoundLike
    {
        /// <summary>
        /// Генерирует таблицу маршрутизации для алгоритма ROU (Round-Robin).
        /// </summary>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// Внешний список представляет собой набор маршрутов, а внутренние списки содержат последовательность узлов (адресов) в маршруте.
        /// </returns>
        /// <remarks>
        /// Алгоритм ROU обычно является простым и может быть полезен для базового сравнения.
        /// </remarks>
        List<List<int>> Generate_ROU_routing();

        /// <summary>
        /// Генерирует таблицу маршрутизации для алгоритма APO (Adaptive Partially Ordered).
        /// </summary>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// </returns>
        /// <remarks>
        /// APO является более сложным алгоритмом учитывающим адаптивность и частичное упорядочивание.
        /// </remarks>
        List<List<int>> Generate_APO_routing();

        /// <summary>
        /// Генерирует таблицу маршрутизации для алгоритма APM (Adaptive Partially Mesh).
        /// </summary>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// </returns>
        /// <remarks>
        /// APM использует адаптивную маршрутизацию и частично связанную топологию (Mesh).
        /// </remarks>
        List<List<int>> Generate_APM_routing();

        /// <summary>
        /// Генерирует таблицу маршрутизации для алгоритма AAO (Adaptive Adaptive Ordered).
        /// </summary>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// </returns>
        /// <remarks>
        /// AAO может подразумевать адаптивную маршрутизацию с учетом порядка.
        /// </remarks>
        List<List<int>> Generate_AAO_routing();

        /// <summary>
        /// Генерирует таблицу маршрутизации для простого статического, алгоритма.
        /// </summary>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// </returns>
        /// <remarks>
        /// Этот метод реализует простейший алгоритм маршрутизации.
        /// </remarks>
        List<List<int>> Generate_Simple_routing();

        /// <summary>
        /// Генерирует таблицу маршрутизации с использованием генетического алгоритма (GA).
        /// </summary>
        /// <param name="token">Токен для отмены операции.</param>
        /// <returns>
        /// Возвращает список списков целых чисел, представляющий таблицу маршрутизации.
        /// </returns>
        /// <remarks>
        /// Этот метод использует генетический алгоритм для поиска оптимальной таблицы маршрутизации.
        /// Использует <see cref="T:System.Threading.CancellationToken"/> для обеспечения возможности отмены долгих вычислений.
        /// </remarks>
        List<List<int>> Generate_GA_routing(CancellationToken token);

        /// <summary>
        /// Вычисляет эффективность для простого алгоритма маршрутизации.
        /// </summary>
        /// <returns>
        /// Возвращает значение типа double, представляющее эффективность.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет эффективность простого алгоритма маршрутизации.
        /// </remarks>
        double Calculate_efficiency_simple();

        /// <summary>
        /// Вычисляет эффективность для алгоритма ROU (Round-Robin).
        /// </summary>
        /// <returns>
        /// Возвращает значение типа double, представляющее эффективность.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет эффективность алгоритма Round-Robin.
        /// </remarks>
        double Calculate_efficiency_ROU();

        /// <summary>
        /// Вычисляет эффективность для алгоритма APM (Adaptive Partially Mesh).
        /// </summary>
        /// <returns>
        /// Возвращает значение типа double, представляющее эффективность.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет эффективность алгоритма APM.
        /// </remarks>
        double Calculate_efficiency_APM();

        /// <summary>
        /// Вычисляет эффективность для алгоритма APO (Adaptive Partially Ordered).
        /// </summary>
        /// <returns>
        /// Возвращает значение типа double, представляющее эффективность.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет эффективность алгоритма APO.
        /// </remarks>
        double Calculate_efficiency_APO();

        /// <summary>
        /// Вычисляет эффективность для генетического алгоритма (GA).
        /// </summary>
        /// <param name="token">Токен для отмены операции.</param>
        /// <returns>
        /// Возвращает значение типа double, представляющее эффективность.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет эффективность генетического алгоритма.
        /// Использует <see cref="T:System.Threading.CancellationToken"/> для обеспечения возможности отмены долгих вычислений.
        /// </remarks>
        double Calculate_efficiency_GA(CancellationToken token);

        /// <summary>
        /// Вычисляет среднее расстояние для генетического алгоритма (GA).
        /// </summary>
        /// <param name="token">Токен для отмены операции.</param>
        /// <returns>
        /// Возвращает значение типа double, представляющее среднее расстояние.
        /// </returns>
        /// <remarks>
        /// Этот метод вычисляет среднее расстояние, пройденное пакетами при использовании генетического алгоритма маршрутизации.
        /// Использует <see cref="T:System.Threading.CancellationToken"/> для обеспечения возможности отмены долгих вычислений.
        /// </remarks>
        double Calculate_average_distance_GA(CancellationToken token);

    }
}